
/* $Header: ScrollLstP.h,v 1.1 89/07/05 15:35:38 hubbard Exp $ */

/*  This is the scrollList widget. It's basically a modification of
 *  Chris Peterson's List widget that allows the list to be viewed
 *  with a scrollbar and to optionally keep the highlighted region
 *  visible. The second feature was what made the modification necessary.
 *  
 *				Jordan Hubbard
 *
 * Based on List widget by:
 *			Chris D. Peterson
 *                      MIT - Project Athena
 *
 * $Log:	ScrollLstP.h,v $
 * Revision 1.1  89/07/05  15:35:38  hubbard
 * Initial revision
 * 
 * 
 */

/***********************************************************
Copyright 1988 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/* 
 * ListP.h - Private definitions for List widget
 * 
 */

#ifndef _XtScrollListP_h
#define _XtScrollListP_h

/***********************************************************************
 *
 * ScrollList Widget Private Data
 *
 ***********************************************************************/

#include "ScrollList.h"
#include "SimpleP.h"

#define NO_HIGHLIGHT            XT_LIST_NONE
#define OUT_OF_RANGE            -1
#define OKAY                     0

#define UP	1
#define DOWN	2

/* New fields for the ScrollList widget class record */

typedef struct {int foo;} ScrollListClassPart;

/* Full class record declaration */
typedef struct _ScrollListClassRec {
    CoreClassPart	core_class;
    SimpleClassPart	simple_class;
    ScrollListClassPart	list_class;
} ScrollListClassRec;

extern ScrollListClassRec scrollListClassRec;

/* New fields for the List widget record */
typedef struct {
    /* resources */
    Pixel	foreground;
    Dimension	internal_width,
        	internal_height,
                column_space,
                row_space;
    int         default_cols;
    Boolean     force_cols,
                paste,
                vertical_cols,
    		repositionOnHilite,
    		scrollVertical;
    int         longest;
    int         nitems;		/* number of items in the list. */
    XFontStruct	*font;
    String *    list;
    XtCallbackList  callback;

    /* private state */

    Widget	sbar;		/* scrollbar widget */
    int         is_highlighted,	/* set to the item currently highlighted. */
                highlight,	/*set to the item that should be highlighted.*/
                col_width,	/* width of each column. */
                row_height,	/* height of each row. */
                nrows,		/* number of rows in the list. */
                ncols,		/* number of columns in the list. */
    		left_margin,	/* left margin in pixels */
    		top,		/* top line visible */
		n_lines,	/* number of lines visible */
    		actual;		/* number of physical lines in list */
    GC		normgc,		/* a couple o' GC's. */
                revgc,
                graygc;		/* used when inactive. */
    Boolean	*stiplist;	/* which item(s) are stippled */

} ScrollListPart;


/****************************************************************
 *
 * Full instance record declaration
 *
 ****************************************************************/

typedef struct _ScrollListRec {
    CorePart		core;
    SimplePart		simple;
    ScrollListPart	list;
} ScrollListRec;

#endif /* _XtScrollListP_h */
